#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "defines.h"

#include "asm_val.h"
#include "asm_field.h"
#include "lib.h"

struct pkt_asm_field_param {
  pkt_asm_field_t field;
  char *key;
  pkt_asm_val_type_t type;
};

#define PKT_ASM_VAL_TYPE_UINT8  PKT_ASM_VAL_TYPE_INTEGER
#define PKT_ASM_VAL_TYPE_UINT16 PKT_ASM_VAL_TYPE_INTEGER
#define PKT_ASM_VAL_TYPE_UINT32 PKT_ASM_VAL_TYPE_INTEGER

static struct pkt_asm_field_param fields[] = {
  { PKT_ASM_FIELD_UNKNOWN_DATA,      "UNKNOWN.DATA",      PKT_ASM_VAL_TYPE_BINARY },

  { PKT_ASM_FIELD_LOOPBACK_FAMILY,   "LOOPBACK.FAMILY",   PKT_ASM_VAL_TYPE_UINT32 },
  { PKT_ASM_FIELD_LOOPBACK_DATA,     "LOOPBACK.DATA",     PKT_ASM_VAL_TYPE_BINARY },

  { PKT_ASM_FIELD_ETHERNET_DST_ADDR, "ETHERNET.DST_ADDR", PKT_ASM_VAL_TYPE_MACADDR },
  { PKT_ASM_FIELD_ETHERNET_SRC_ADDR, "ETHERNET.SRC_ADDR", PKT_ASM_VAL_TYPE_MACADDR },
  { PKT_ASM_FIELD_ETHERNET_TYPE,     "ETHERNET.TYPE",     PKT_ASM_VAL_TYPE_UINT16 },
  { PKT_ASM_FIELD_ETHERNET_VLAN_TAG, "ETHERNET.VLAN_TAG", PKT_ASM_VAL_TYPE_UINT16 },
  { PKT_ASM_FIELD_ETHERNET_DATA,     "ETHERNET.DATA",     PKT_ASM_VAL_TYPE_BINARY },

  { PKT_ASM_FIELD_ARP_HARDWARE,       "ARP.HARDWARE",       PKT_ASM_VAL_TYPE_UINT16 },
  { PKT_ASM_FIELD_ARP_PROTOCOL,       "ARP.PROTOCOL",       PKT_ASM_VAL_TYPE_UINT16 },
  { PKT_ASM_FIELD_ARP_HARDWARE_SIZE,  "ARP.HARDWARE_SIZE",  PKT_ASM_VAL_TYPE_UINT8 },
  { PKT_ASM_FIELD_ARP_PROTOCOL_SIZE,  "ARP.PROTOCOL_SIZE",  PKT_ASM_VAL_TYPE_UINT8 },
  { PKT_ASM_FIELD_ARP_OPERATION,      "ARP.OPERATION",      PKT_ASM_VAL_TYPE_UINT16 },
  { PKT_ASM_FIELD_ARP_SENDER_MACADDR, "ARP.SENDER_MACADDR", PKT_ASM_VAL_TYPE_MACADDR },
  { PKT_ASM_FIELD_ARP_SENDER_IPADDR,  "ARP.SENDER_IPADDR",  PKT_ASM_VAL_TYPE_IPADDR },
  { PKT_ASM_FIELD_ARP_TARGET_MACADDR, "ARP.TARGET_MACADDR", PKT_ASM_VAL_TYPE_MACADDR },
  { PKT_ASM_FIELD_ARP_TARGET_IPADDR,  "ARP.TARGET_IPADDR",  PKT_ASM_VAL_TYPE_IPADDR },

  { PKT_ASM_FIELD_ICMP_TYPE,     "ICMP.TYPE",     PKT_ASM_VAL_TYPE_UINT8 },
  { PKT_ASM_FIELD_ICMP_CODE,     "ICMP.CODE",     PKT_ASM_VAL_TYPE_UINT8 },
  { PKT_ASM_FIELD_ICMP_CHECKSUM, "ICMP.CHECKSUM", PKT_ASM_VAL_TYPE_UINT16 },
  { PKT_ASM_FIELD_ICMP_DATA,     "ICMP.DATA",     PKT_ASM_VAL_TYPE_BINARY },

  { PKT_ASM_FIELD_IGMP_TYPE,     "IGMP.TYPE",     PKT_ASM_VAL_TYPE_UINT8 },
  { PKT_ASM_FIELD_IGMP_CODE,     "IGMP.CODE",     PKT_ASM_VAL_TYPE_UINT8 },
  { PKT_ASM_FIELD_IGMP_CHECKSUM, "IGMP.CHECKSUM", PKT_ASM_VAL_TYPE_UINT16 },
  { PKT_ASM_FIELD_IGMP_GROUP,    "IGMP.GROUP",    PKT_ASM_VAL_TYPE_IPADDR },
  { PKT_ASM_FIELD_IGMP_DATA,     "IGMP.DATA",     PKT_ASM_VAL_TYPE_BINARY },

  { PKT_ASM_FIELD_TCP_SRC_PORT,   "TCP.SRC_PORT",   PKT_ASM_VAL_TYPE_UINT16 },
  { PKT_ASM_FIELD_TCP_DST_PORT,   "TCP.DST_PORT",   PKT_ASM_VAL_TYPE_UINT16 },
  { PKT_ASM_FIELD_TCP_SEQ_NUMBER, "TCP.SEQ_NUMBER", PKT_ASM_VAL_TYPE_UINT32 },
  { PKT_ASM_FIELD_TCP_ACK_NUMBER, "TCP.ACK_NUMBER", PKT_ASM_VAL_TYPE_UINT32 },
  { PKT_ASM_FIELD_TCP_OFFSET,     "TCP.OFFSET",     PKT_ASM_VAL_TYPE_UINT8 },
  { PKT_ASM_FIELD_TCP_FLAGS,      "TCP.FLAGS",      PKT_ASM_VAL_TYPE_UINT8 },
  { PKT_ASM_FIELD_TCP_WINDOW,     "TCP.WINDOW",     PKT_ASM_VAL_TYPE_UINT16 },
  { PKT_ASM_FIELD_TCP_CHECKSUM,   "TCP.CHECKSUM",   PKT_ASM_VAL_TYPE_UINT16 },
  { PKT_ASM_FIELD_TCP_URGENT,     "TCP.URGENT",     PKT_ASM_VAL_TYPE_UINT16 },
  { PKT_ASM_FIELD_TCP_DATA,       "TCP.DATA",       PKT_ASM_VAL_TYPE_BINARY },

  { PKT_ASM_FIELD_UDP_SRC_PORT, "UDP.SRC_PORT", PKT_ASM_VAL_TYPE_UINT16 },
  { PKT_ASM_FIELD_UDP_DST_PORT, "UDP.DST_PORT", PKT_ASM_VAL_TYPE_UINT16 },
  { PKT_ASM_FIELD_UDP_SIZE,     "UDP.SIZE",     PKT_ASM_VAL_TYPE_UINT16 },
  { PKT_ASM_FIELD_UDP_CHECKSUM, "UDP.CHECKSUM", PKT_ASM_VAL_TYPE_UINT16 },
  { PKT_ASM_FIELD_UDP_DATA,     "UDP.DATA",     PKT_ASM_VAL_TYPE_BINARY },

  { PKT_ASM_FIELD_IP_VERSION,     "IP.VERSION",     PKT_ASM_VAL_TYPE_UINT8 },
  { PKT_ASM_FIELD_IP_HEADER_SIZE, "IP.HEADER_SIZE", PKT_ASM_VAL_TYPE_UINT8 },
  { PKT_ASM_FIELD_IP_TOS,         "IP.TOS",         PKT_ASM_VAL_TYPE_UINT8 },
  { PKT_ASM_FIELD_IP_TOTAL_SIZE,  "IP.TOTAL_SIZE",  PKT_ASM_VAL_TYPE_UINT16 },
  { PKT_ASM_FIELD_IP_ID,          "IP.ID",          PKT_ASM_VAL_TYPE_UINT16 },
  { PKT_ASM_FIELD_IP_FRAGMENT,    "IP.FRAGMENT",    PKT_ASM_VAL_TYPE_UINT16 },
  { PKT_ASM_FIELD_IP_TTL,         "IP.TTL",         PKT_ASM_VAL_TYPE_UINT8 },
  { PKT_ASM_FIELD_IP_PROTOCOL,    "IP.PROTOCOL",    PKT_ASM_VAL_TYPE_UINT8 },
  { PKT_ASM_FIELD_IP_CHECKSUM,    "IP.CHECKSUM",    PKT_ASM_VAL_TYPE_UINT16 },
  { PKT_ASM_FIELD_IP_SRC_ADDR,    "IP.SRC_ADDR",    PKT_ASM_VAL_TYPE_IPADDR },
  { PKT_ASM_FIELD_IP_DST_ADDR,    "IP.DST_ADDR",    PKT_ASM_VAL_TYPE_IPADDR },
  { PKT_ASM_FIELD_IP_DATA,        "IP.DATA",        PKT_ASM_VAL_TYPE_BINARY },

  { PKT_ASM_FIELD_ICMPV6_TYPE,     "ICMPV6.TYPE",     PKT_ASM_VAL_TYPE_UINT8 },
  { PKT_ASM_FIELD_ICMPV6_CODE,     "ICMPV6.CODE",     PKT_ASM_VAL_TYPE_UINT8 },
  { PKT_ASM_FIELD_ICMPV6_CHECKSUM, "ICMPV6.CHECKSUM", PKT_ASM_VAL_TYPE_UINT16 },
  { PKT_ASM_FIELD_ICMPV6_DATA,     "ICMPV6.DATA",     PKT_ASM_VAL_TYPE_UINT32 },

  { PKT_ASM_FIELD_IPV6_VERSION,      "IPV6.VERSION",      PKT_ASM_VAL_TYPE_UINT8 },
  { PKT_ASM_FIELD_IPV6_FLOWID,       "IPV6.FLOWID",       PKT_ASM_VAL_TYPE_UINT32 },
  { PKT_ASM_FIELD_IPV6_PAYLOAD_SIZE, "IPV6.PAYLOAD_SIZE", PKT_ASM_VAL_TYPE_UINT16 },
  { PKT_ASM_FIELD_IPV6_NEXT_HEADER,  "IPV6.NEXT_HEADER",  PKT_ASM_VAL_TYPE_UINT8 },
  { PKT_ASM_FIELD_IPV6_HOP_LIMIT,    "IPV6.HOP_LIMIT",    PKT_ASM_VAL_TYPE_UINT8 },
  { PKT_ASM_FIELD_IPV6_SRC_ADDR,     "IPV6.SRC_ADDR",     PKT_ASM_VAL_TYPE_IPV6ADDR },
  { PKT_ASM_FIELD_IPV6_DST_ADDR,     "IPV6.DST_ADDR",     PKT_ASM_VAL_TYPE_IPV6ADDR },
  { PKT_ASM_FIELD_IPV6_EXT_SIZE,     "IPV6.EXT_SIZE",     PKT_ASM_VAL_TYPE_UINT8 },
  { PKT_ASM_FIELD_IPV6_EXT_DATA,     "IPV6.EXT_DATA",     PKT_ASM_VAL_TYPE_BINARY },
  { PKT_ASM_FIELD_IPV6_DATA,         "IPV6.DATA",         PKT_ASM_VAL_TYPE_BINARY },

  { PKT_ASM_FIELD_NONE, NULL, PKT_ASM_VAL_TYPE_NONE }
};

static struct pkt_asm_field_param *search_field(pkt_asm_field_t field)
{
  struct pkt_asm_field_param *p;
  for (p = fields; p->field != PKT_ASM_FIELD_NONE; p++) {
    if (p->field == field)
      break;
  }
  return p;
}

static struct pkt_asm_field_param *search_key(char *key)
{
  struct pkt_asm_field_param *p;
  for (p = fields; p->field != PKT_ASM_FIELD_NONE; p++) {
    if (!strcmp(p->key, key))
      break;
  }
  return p;
}

char *pkt_asm_field_get_key(pkt_asm_field_t field)
{
  struct pkt_asm_field_param *p;
  p = search_field(field);
  if (p) return p->key;
  return NULL;
}

pkt_asm_val_type_t pkt_asm_field_get_type(pkt_asm_field_t field)
{
  struct pkt_asm_field_param *p;
  p = search_field(field);
  if (p) return p->type;
  return PKT_ASM_VAL_TYPE_NONE;
}

pkt_asm_field_t pkt_asm_field_search_field_from_key(char *key)
{
  struct pkt_asm_field_param *p;
  p = search_key(key);
  if (p) return p->field;
  return PKT_ASM_FIELD_NONE;
}

int pkt_asm_field_output_key_list(FILE *fp, char *header)
{
  struct pkt_asm_field_param *p;
  for (p = fields; p->field != PKT_ASM_FIELD_NONE; p++) {
    if (header)
      fprintf(fp, "%s", header);
    fprintf(fp, "%s\n", p->key);
  }
  return 0;
}
