#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>
#ifndef WIN32
#include <sys/socket.h>
#include <netinet/in.h>
#ifdef USE_GETHOSTBYNAME
#include <netdb.h>
#else
#include <arpa/inet.h>
#endif
#else
#include <fcntl.h>
#include <winsock2.h>
#include <ws2tcpip.h>
#endif

#ifdef WIN32
#define USE_OLDSIGNAL
#ifndef USE_GETHOSTBYNAME
#ifndef USE_NETLIB
#warning Cannot use inet_pton() on Windows?
#warning If an error as "undefined reference to inet_pton" occurred
#warning on the link stage, Please try to use netlib as follows.
#warning $ unzip netlib-X.X.zip ; mv netlib-X.X netlib ; make USE_NETLIB=yes
#else
int inet_pton(int af, const char *src, void *dst);
#endif
#endif
#endif

#define BUFFER_SIZE (1024*80)

#define LOG(args...) do { fprintf(stderr, args); } while (0)
#define VLOG(args...) do { if (verbose) fprintf(stderr, args); } while (0)
#define ERROR_EXIT(args...) do { fprintf(stderr, args); exit(1); } while (0)
#define GOTO_DONE(args...) do { fprintf(stderr, args); goto done; } while (0)

static void help()
{
  printf("tcp6-recv [-v] [<hostname>] <port>\n");
  exit(0);
}

static int terminated = 0;
static void sigint_handler(int value)
{
  terminated = 1;
}

#ifdef WIN32
int winsock_init()
{
  WSADATA wsaData;
  if (WSAStartup(MAKEWORD(2, 0), &wsaData) != 0)
    ERROR_EXIT("WSAStartup() failed.\n");
  return 0;
}
#endif

int main(int argc, char *argv[])
{
  char *hostname = NULL;
  int verbose = 0, port, s, accept_s, recvsize, writesize;
  int total_recvsize = 0, total_writesize = 0;
  struct sockaddr_in6 addr;
  socklen_t addrlen;
  char buffer[BUFFER_SIZE];
#ifdef USE_GETHOSTBYNAME
  struct hostent *host = NULL;
#endif
#ifndef USE_OLDSIGNAL
  struct sigaction sa;
#endif
#ifndef WIN32
  int on = 1;
#else
  BOOL on = TRUE;
#endif

  argc--; argv++;

  if ((argc > 0) && !strcmp(argv[0], "-v")) {
    verbose = 1;
    argc--; argv++;
  }

  if (argc < 1)
    help();

  if (argc > 1) {
    hostname = *argv;
    argc--; argv++;
  }
  port = atoi(*argv);
  argc--; argv++;

#ifdef WIN32
  winsock_init();
  if (!isatty(1))
    setmode(1, O_BINARY);
#endif

#ifdef USE_GETHOSTBYNAME
  if (hostname) {
    /* ۥ̾饢ɥ쥹 */
    host = gethostbyname(hostname);
    if (host == NULL)
      ERROR_EXIT("Cannot resolve hostname.\n");
  }
#endif

  /* åȤΥץ */
  s = socket(PF_INET6, SOCK_STREAM, 0);
  if (s < 0)
    ERROR_EXIT("Cannot open socket.\n");

  /* ɥ쥹Ѳǽˤ */
#ifndef WIN32
  if (setsockopt(s, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on)) < 0)
#else
  if (setsockopt(s, SOL_SOCKET, SO_REUSEADDR, (char *)&on, sizeof(on)) < 0)
#endif
    ERROR_EXIT("Cannot set socket option.\n");

  /* ɥ쥹 */
  memset(&addr, 0, sizeof(addr));
  addr.sin6_family = AF_INET6;
  addr.sin6_port = htons(port);
  if (hostname) {
    /* Υɥ쥹Ǥ³Τ߼դ */
#ifdef USE_GETHOSTBYNAME
    memcpy(&addr.sin6_addr, host->h_addr, host->h_length);
#else
    if (inet_pton(AF_INET6, hostname, &addr.sin6_addr) < 1)
      ERROR_EXIT("Invalid address.\n");
#endif
  } else {
    /* ٤ƤΥɥ쥹Ǥ³դ */
    memcpy(&addr.sin6_addr, &in6addr_any, sizeof(addr.sin6_addr));
  }
  addrlen = sizeof(addr);

  /* ɥ쥹 */
  if (bind(s, (struct sockaddr *)&addr, addrlen) < 0)
    ERROR_EXIT("Cannot bind.\n");

  /* ³塼ξ¤ */
  if (listen(s, 5) < 0)
    ERROR_EXIT("Cannot listen.\n");

  /* 饤Ȥ³Ԥ */
  accept_s = s;
  memset(&addr, 0, sizeof(addr));
  addrlen = sizeof(addr);
  s = accept(accept_s, (struct sockaddr *)&addr, &addrlen);
  if (s < 0)
    ERROR_EXIT("Cannot accept.\n");

  /* ʥϥɥϿ */
#ifndef USE_OLDSIGNAL
  memset(&sa, 0, sizeof(sa));
  sa.sa_handler = sigint_handler;
  if (sigaction(SIGINT, &sa, NULL) < 0)
#else
  if (signal(SIGINT, sigint_handler) == SIG_ERR)
#endif
    ERROR_EXIT("Cannot set signal.\n");

  while (!terminated) {
    /* ǡμ */
#ifndef WIN32
    recvsize = read(s, buffer, sizeof(buffer));
#else
    recvsize = recv(s, buffer, sizeof(buffer), 0);
#endif
    if (recvsize < 0)
      GOTO_DONE("Cannot recv.\n");
    if (recvsize == 0)
      break;

    VLOG("Recv size on TCP6:\t%d\n", recvsize);
    total_recvsize += recvsize;

    /* ɸϤ˥ǡ񤭽Ф */
    writesize = write(1, buffer, recvsize);
    if (writesize < 0)
      GOTO_DONE("Cannot write.\n");
    VLOG("\n");

    VLOG("Write size to stdout:\t%d\n", writesize);
    total_writesize += writesize;
  }

done:
  if (terminated)
    LOG("Terminated.\n");

  VLOG("\n");
  VLOG("Total recv size:\t%d\n", total_recvsize);
  VLOG("Total write size:\t%d\n", total_writesize);

  /* åȤΥ */
#ifndef WIN32
  shutdown(s, SHUT_RDWR);
  close(s);
  close(accept_s);
#else
  shutdown(s, SD_BOTH);
  closesocket(s);
  closesocket(accept_s);
#endif

  return 0;
}
